# blroots
# Copyright (C) 2025 blroots Project Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Implements various key caches on instances of Blender objects.

In particular,

"""

import functools
import typing as typ

# import griffe
from .bl_id_props import BLIDProps
from .bl_prop import BLProp, SupportedBLPropType


####################
# - BLField
####################
class BLField:
	"""A descriptor that allows persisting arbitrary types in Blender objects, with cached reads."""

	def __init__(
		self,
		default_value: typ.Any | None = None,
		*,
		display_name: str | None = None,
		description: str | None = None,
		**bl_prop_info: dict[str, typ.Any],
	) -> None:
		"""Initializes and sets the attribute to a given default value.

		The attribute **must** declare a type annotation, and it **must** match the type of `default_value`.

		Parameters:
			default_value: The default value to use if the value is read before it's set.
		"""
		self.default_value: typ.Any | None = default_value
		self.bl_prop_info: dict[str, typ.Any] = bl_prop_info

		self.name: str | None = None
		self.display_name: str | None = display_name
		self.description: str | None = description
		self.type_hint: type | None = None

	####################
	# - Initialization
	####################
	def __set_name__(self, owner: type[SupportedBLPropType], name: str) -> None:
		"""Sets up this descriptor on the class, preparing it for per-instance use.

		A `BLProp` is constructed using this descriptor's attribute name on `owner`, and the `self.prop_info` previously created during `self.__init__()`.
		Then, a corresponding / underlying `bpy.types.Property` is initialized on `owner` using `self.bl_prop.init_bl_type(owner)`

		Notes:
			Run by Python when setting an instance of a "descriptor" class, to an attribute of another class (denoted `owner`).
			For more, search for "Python descriptor protocol".

		Parameters:
			owner: The class that contains an attribute assigned to an instance of this descriptor.
			name: The name of the attribute that an instance of descriptor was assigned to.
		"""
		####################
		# - Parse Information for BLProp
		####################
		## *self.bl_prop is only valid to use after this section runs.
		if issubclass(
			owner,
			BLIDProps | SupportedBLPropType,
		):
			# Parse BLProp from Type Annotation
			## - TODO: Handle errors to clarify incorrect bl_prop_info stuff.
			self.name = name
			self.display_name = (
				self.display_name
				if self.display_name is not None
				else self.name.replace('-', ' ').replace('_', ' ').title()
			)
			self.type_hint = typ.get_type_hints(owner, include_extras=True).get(name)

			# Parse Description
			if self.description is None:
				# Try Parsing from owner.__doc__
				## - TODO: Parse __doc__ of the bl_class for property descriptions.
				# if hasattr(owner, '__doc__') and isinstance(owner.__doc__, str):
				# docstring_sections = griffe.Docstring(owner.__doc__, lineno=1).parse('google')

				# if 'parameters' in docstring_sections:
				# idx = docstring_sections.index('parameters')
				# docstring_sections[idx].as_dict()

				# if name in [param for param in docstring.params]
				self.description = ''
		else:
			msg = f"Cannot declare 'BLField' (name='{name}') on unsupported class '{owner}'."
			raise TypeError(msg)

		####################
		# - Initialize Owner
		####################
		## This is only done if the owner should  directly accessed at this point.
		## Ex. BLIDProps has its own register() method.
		if issubclass(owner, SupportedBLPropType):
			self.bl_prop.init_on_bl_class(owner)

	@functools.cached_property
	def bl_prop(self) -> BLProp:
		"""Object used for convenient interaction with the Blender class.

		Notes:
			Since `BLProp` keeps no state about the field's class.

		Warnings:
			**MUST** be called for the first time after `self.type_hint` is set within `self.__set_name__`.

			Additionally, within `self.__set_name__`, `self.bl_prop.init_on_bl_class(owner)` MUST be called on the field's class.
			This effectively synchronizes the relevant field on the class with this `BLProp`.
		"""
		if (
			self.type_hint is not None
			and self.name is not None
			and self.display_name is not None
			and self.description is not None
		):
			return BLProp.from_type_hint(
				self.type_hint,
				name=self.name,
				display_name=self.display_name,
				description=self.description,
				default_value=self.default_value,
				bl_prop_info=self.bl_prop_info,
			)
		msg = "'BLField.bl_prop' was called before the descriptor was initialized in '__set_name__'."
		raise ValueError(msg)

	####################
	# - Descriptor __get__/__set__
	####################
	def __get__(
		self,
		bl_instance: SupportedBLPropType | None,
		owner: type[SupportedBLPropType],
	) -> typ.Any:
		"""Retrieves the value described by the BLField.

		Notes:
			Run by Python when the attribute described by the descriptor is accessed.
			For more, search for "Python descriptor protocol".

		Parameters:
			bl_instance: Instance that is accessing the attribute.
			owner: The class that owns the instance.
		"""
		return self.bl_prop.read(bl_instance)

	def __set__(self, bl_instance: SupportedBLPropType | None, value: typ.Any) -> None:
		"""Sets the value described by the BLField.

		In general, any BLField modified in the UI will set `InvalidateCache` on this descriptor.
		If `self.prop_info['use_prop_update']` is set, the method `bl_instance.on_prop_changed(self.bl_prop.name)` will then be called and start a `FlowEvent.DataChanged` event chain.

		Notes:
			Run by Python when the attribute described by the descriptor is set.
			For more, search for "Python descriptor protocol".

		Parameters:
			bl_instance: Instance that is accessing the attribute.
			owner: The class that owns the instance.
		"""
		self.bl_prop.write(bl_instance, value)
