# blroots
# Copyright (C) 2025 blroots Project Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Implements `ClassProp`."""

import functools
import inspect
import typing as typ

import bpy

from blroots.utils import bl_instance, logger

from ..bl_prop import BLProp, bl_prop_instance_property_name

log = logger.get(__name__)

PRIORITY = 11

## TODO: Main thing, I think, at this point, is to find a way to register those ProeprtyGroups mindfully, then go play with it!


@typ.runtime_checkable
class BLPropClass(typ.Protocol):
	@classmethod
	def bl_prop_group(cls) -> type[bpy.types.PropertyGroup]: ...

	## TODO: The bl_prop_group should be checked for:
	## - That it is a pydantic BaseModel.
	## - That the BaseModel has frozen=True.
	## - That data attributes are identical between the BL_ class and this version.
	## - That type annotations on data attributes are identical.
	## TODO: USAGE
	## - This class can be used freely without a backing BL_ class.
	## - This class can also be used to type-annotate a BLField.
	## -- The BL_ class must be registered before being assigned to anything.
	## -- Since frozen=True, only the entire instance can be replaced at a time.
	## --- Underneath, a backing PropertyGroup will be mutated as needed.
	## -- bl_events must be propagated to the backing PropertyGroup.
	## --- Use bl_event.propagate with source='parent' to get data from the PropertyGroup's parent.
	## --- Let's make sure to make it easy for instances of PropertyGroup to get that parent!
	## --- We can later do something similar for data pipelines between nodes and sockets.


# def to_bl_prop_group(
# self, bl_prop_group: BL_LoggingPrefs
# ) -> type[bpy.types.PropertyGroup]:
# bl_prop_group.use_log_file = self.use_log_file
# bl_prop_group.log_file_level = self.log_file_level
# bl_prop_group.log_file_path = self.log_file_path
#
# bl_prop_group.use_log_console = self.use_log_console
# bl_prop_group.log_console_level = self.log_console_level
#
# @classmethod
# def from_bl_prop_group(cls, bl_prop_group: BL_LoggingPrefs) -> typ.Self:
# return cls(
# use_log_file=bl_prop_group.use_log_file,
# log_file_level=bl_prop_group.log_file_level,
# log_file_path=bl_prop_group.log_file_path,
# use_log_console=bl_prop_group.use_log_console,
# log_console_level=bl_prop_group.log_console_level,
# )


class ClassProp(BLProp, frozen=True):
	"""A single class; in particular, an appropriately configured `pydantic` BaseModel."""

	use_property_group = True

	default_value: None

	####################
	# - Type Support
	####################
	@classmethod
	def supports_type_hint(cls, field_type_hint: type) -> bool:
		"""Support classes that conform to the `BLPropClass` protocol."""
		return inspect.isclass(field_type_hint) and isinstance(
			field_type_hint, BLPropClass
		)

	@property
	def bpy_property_group(self) -> type[bpy.props.PropertyGroup]:
		"""Dynamically create a PropertyGroup."""
		return self.type_hint.bl_prop_group()
		## TODO: Who registers it, and when?
		## TODO: I don't like having the "two-class" paradigm.
		## - On the plus side, "outside classes" can be adapted.
		## - On the minus side, it's two things now. Two parameter documentations to parse, etc. .
		## - If we did go with an all-in-one, typ.Annotated w/partial BLField constructor?
		## - Or somehow get pydantic to understand setting it to a BLField?
		## - I honestly don't know what to do exactly...
		## - On the other hand hand, look how mindful and simple this method is.

	####################
	# - bpy.props Interface
	####################
	@property
	def bpy_prop_type(self) -> type[bpy.props.PointerProperty]:
		"""Use `bpy.props.PointerProperty` to reference a particular pre-made `bpy.types.PropertyGroup` subclass."""
		return bpy.props.PointerProperty

	@functools.cached_property
	def bpy_prop_kwargs(self) -> dict[str, typ.Any]:
		"""Keyword arguments to pass to `self.bpy_prop_type`."""
		return {
			'type': self.bpy_property_group,
		}

	####################
	# - Encode/Decode
	####################
	def encode(self, value: typ.Any) -> dict[str, typ.Any]:
		"""Encode a value to a dictionary that will be written to a `PropertyGroup`.

		Notes:
			The pickle is this:

			- On the PropertyGroup, each field is associated with a unique BLProp instance.
			- During BLProp initialization, a class method is created to allow accessing the BLProp underlying any BLField.
			- Why do we need to do this? BLProp instance uses encode()/decode() to translate between user-facing and Blender-facing representations of data.
			- Therefore,for each subproperty, retrieve the `BLProp` instance, then use its `.encode()` method to transform `value` fields down to the Blender representation that will be expected by the `PropertyGroup`.
			- This naturally provides correct recursion, too, since `.encode()` will simply chain its way all the way down to the bottom.

		"""
		if isinstance(value, self.type_hint):
			return {
				bl_prop_name: (
					bl_prop := getattr(  # noqa: F841
						self.bpy_property_group,
						bl_prop_instance_property_name(bl_prop_name),
					)()
				).encode(getattr(value, bl_prop_name))
				for bl_prop_name in self.bpy_property_group.bl_props
			}

		raise NotImplementedError

	def decode(self, raw_bl_value: bpy.types.PropertyGroup) -> typ.Any:
		"""Decode the subclass of `bpy.types.PropertyGroup` subclass into the `pydantic.BaseModel` contained within `self.type_hint`.

		Notes:
			Simply reading the `BLField`s invokes `.decode()` recursively and correctly.
			Thus, decoding ends up being comparatively simpler than `encode()`.
		"""
		if isinstance(raw_bl_value, self.bpy_property_group):
			return self.type_hint(
				**{
					bl_prop_name: getattr(raw_bl_value, bl_prop_name)
					for bl_prop_name in self.bpy_property_group.bl_props
				}
			)

		raise NotImplementedError

	####################
	# - UI
	####################
	def draw(
		self,
		bl_instance: bl_instance.BLInstance,
		layout: bpy.types.UILayout,
		_: bpy.types.Context,
	) -> None:
		"""Draw this property to the given Blender `UILayout`."""
		prop_group = getattr(bl_instance, self.bl_name)
		self.type_hint.__class__.draw(prop_group)

	####################
	# - Selection
	####################
	@classmethod
	def priority(cls) -> int:
		"""Priority with which to prefer this implementation over others compatible with the same type hint."""
		return PRIORITY
