# blroots
# Copyright (C) 2025 blroots Project Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Implements `PathProp`."""

import builtins
import functools
import typing as typ
from pathlib import Path

import bpy

from ..bl_prop import BLProp, SupportedBLPropType

PRIORITY = 20


class PathProp(BLProp, frozen=True):
	"""A single constrained floating-point number.

	Parameters:
		abs_min: The absolute minimum value of the floating point number.
		abs_max: The absolute maximum value of the floating point number.
		soft_min: The lowest value that can be set using the UI slider.
		soft_max: The highest value that can be set using the UI slider.
		float_prec: The number of decimal places displayed in the UI.
		float_step: The size of each mouse-move increment when using the UI slider.
			_Divide this integer by `100` for actual change per increment._
	"""

	default_value: Path
	## TODO: Use 'before' field validator to select a default value when None is specified, then an 'after' field validator to constrain it to abs_min and abs_max.

	# Constraints
	is_file: bool = False
	is_dir: bool = False
	## TODO: Validator that disallows both is_file and is_dir

	## TODO: A "relative" constraint that uses '//'

	####################
	# - Type Support
	####################
	@classmethod
	def supports_type_hint(cls, field_type_hint: type) -> bool:
		"""Support only the explicit type `float` aka. `builtins.float`."""
		return field_type_hint is Path

	####################
	# - bpy.props Interface
	####################
	@property
	def bpy_prop_type(self) -> type[bpy.props.StringProperty]:
		"""The underlying `bpy.props` class backing this property."""
		return bpy.props.StringProperty  # type: ignore[no-any-return]

	@functools.cached_property
	def bpy_prop_kwargs(self) -> dict[str, typ.Any]:
		"""Keyword arguments to pass to `self.bpy_prop_type`."""
		kwargs: dict[str, typ.Any] = {
			'default': self.encode(self.default_value),
		}

		if self.is_file:
			kwargs |= {'subtype': 'FILE_PATH'}

		elif self.is_dir:
			kwargs |= {'subtype': 'DIR_PATH'}

		else:
			kwargs |= {'subtype': 'FILE_PATH'}

		return kwargs

	####################
	# - Encode/Decode
	####################
	def encode(self, value: typ.Any) -> str:
		"""Encode a value that supports `float()` to a Blender float."""
		# Parse the Value
		if isinstance(value, Path):
			parsed_value = str(value.resolve())
		else:
			msg = f"Value '{value}' must be a Path."
			raise TypeError(msg)

		return parsed_value

	def decode(self, raw_bl_value: str) -> Path:
		"""Decode a Blender float by doing nothing."""
		return Path(bpy.path.abspath(raw_bl_value))

	####################
	# - UI
	####################
	def draw(
		self,
		bl_instance: SupportedBLPropType,
		layout: bpy.types.UILayout,
		text: str | None = None,
		placeholder: str = '',
	) -> None:
		"""Draw this property to the given Blender `UILayout`."""
		layout.prop(
			bl_instance,
			self.bl_name,
			text=self.display_name if text is None else text,
			placeholder=placeholder,
		)

	####################
	# - Selection
	####################
	@classmethod
	def priority(cls) -> int:
		"""Priority with which to prefer this implementation over others compatible with the same type hint."""
		return PRIORITY
