# blroots
# Copyright (C) 2025 blroots Project Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Implements `StrEnumProp`."""

import enum
import functools
import typing as typ

import bpy

from ..bl_prop import BLProp, SupportedBLPropType

PRIORITY = 5


class StrEnumProp(BLProp, frozen=True):
	"""A static enumeration."""

	default_value: enum.StrEnum
	## TODO: Use 'before' field validator to select a default value when None is specified, then an 'after' field validator to constrain it to abs_min and abs_max.

	####################
	# - Type Support
	####################
	@classmethod
	def supports_type_hint(cls, field_type_hint: type) -> bool:
		"""Support only the explicit type `float` aka. `builtins.float`."""
		return (
			issubclass(field_type_hint, enum.StrEnum)
			and hasattr(field_type_hint, 'bl_display_name')
			and hasattr(field_type_hint, 'bl_icon')
		)

	####################
	# - bpy.props Interface
	####################
	@property
	def bpy_prop_type(self) -> type[bpy.props.EnumProperty]:
		"""The underlying `bpy.props` class backing this property."""
		return bpy.props.EnumProperty  # type: ignore[no-any-return]

	@functools.cached_property
	def bpy_prop_kwargs(self) -> dict[str, typ.Any]:
		"""Keyword arguments to pass to `self.bpy_prop_type`."""
		kwargs: dict[str, typ.Any] = {
			'items': [
				(
					el.name,
					el.bl_display_name,
					el.bl_display_name,
					el.bl_icon,
					idx,
				)
				for idx, el in enumerate(self.type_hint)
			],
			'default': self.encode(self.default_value),
			## TODO: Parse Enum.__doc__ to get description from the Attributes section.
		}

		return kwargs

	####################
	# - Encode/Decode
	####################
	def encode(self, value: typ.Any) -> enum.StrEnum:
		"""Encode a value that supports `float()` to a Blender float."""
		if isinstance(value, self.type_hint):
			parsed_value = value

		elif isinstance(value, str):
			parsed_value = getattr(self.type_hint, value, None)
			if parsed_value is None:
				msg = f"Object '{value}' could not be coerced to the enum {self.type_hint} during encoding."
				raise ValueError(msg)

		return parsed_value.name

	def decode(self, raw_bl_value: str) -> str:
		"""Decode a Blender float by doing nothing."""
		parsed_value = getattr(self.type_hint, raw_bl_value, None)
		if parsed_value is not None:
			return raw_bl_value

		msg = f"String '{raw_bl_value}' could not be coerced to the enum {self.type_hint} during decoding."
		raise ValueError(msg)

	####################
	# - UI
	####################
	def draw(
		self,
		bl_instance: SupportedBLPropType,
		layout: bpy.types.UILayout,
		text: str | None = None,
	) -> None:
		"""Draw this property to the given Blender `UILayout`."""
		layout.prop(
			bl_instance,
			self.bl_name,
			text=self.display_name if text is None else text,
		)

	####################
	# - Selection
	####################
	@classmethod
	def priority(cls) -> int:
		"""Priority with which to prefer this implementation over others compatible with the same type hint."""
		return PRIORITY
