# blroots
# Copyright (C) 2025 blroots Project Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Implements `StrProp`."""

import builtins
import functools
import typing as typ

import bpy

from ..bl_prop import BLProp, SupportedBLPropType

PRIORITY = 20


class StrProp(BLProp, frozen=True):
	"""A single constrained string.

	Parameters:
		is_secret: Whether the string should be considered a secret value.
			Secrets display as dots in the interface, and aren't saved in the `.blend` on save.
	"""

	default_value: str
	## TODO: Use 'before' field validator to select a default value when None is specified, then an 'after' field validator to constrain it to abs_min and abs_max.

	# Constraints
	is_secret: bool = False
	## TODO: Search Callback
	## TODO: Regex Validator w/UI Feedback?

	####################
	# - Type Support
	####################
	@classmethod
	def supports_type_hint(cls, field_type_hint: type) -> bool:
		"""Support only the explicit type `float` aka. `builtins.float`."""
		return field_type_hint is builtins.str

	####################
	# - bpy.props Interface
	####################
	@property
	def bpy_prop_type(self) -> type[bpy.props.StringProperty]:
		"""The underlying `bpy.props` class backing this property."""
		return bpy.props.StringProperty  # type: ignore[no-any-return]

	@functools.cached_property
	def bpy_prop_kwargs(self) -> dict[str, typ.Any]:
		"""Keyword arguments to pass to `self.bpy_prop_type`."""
		kwargs: dict[str, typ.Any] = {
			'default': self.default_value,
		}

		if self.is_secret is not None:
			kwargs |= {'subtype': 'PASSWORD', 'options': ['SKIP_SAVE']}

		return kwargs

	####################
	# - Encode/Decode
	####################
	def encode(self, value: typ.Any) -> str:
		"""Encode a value that supports `float()` to a Blender float."""
		# Parse the Value
		try:
			parsed_value = str(value)
		except ValueError as ex:
			msg = f"Value '{value}' could not be coerced to a float."
			raise ValueError(msg) from ex

		return parsed_value

	def decode(self, raw_bl_value: str) -> str:
		"""Decode a Blender float by doing nothing."""
		return raw_bl_value

	####################
	# - UI
	####################
	def draw(
		self,
		bl_instance: SupportedBLPropType,
		layout: bpy.types.UILayout,
		text: str | None = None,
		placeholder: str = '',
	) -> None:
		"""Draw this property to the given Blender `UILayout`."""
		layout.prop(
			bl_instance,
			self.bl_name,
			text=self.display_name if text is None else text,
			placeholder=placeholder,
		)

	####################
	# - Selection
	####################
	@classmethod
	def priority(cls) -> int:
		"""Priority with which to prefer this implementation over others compatible with the same type hint."""
		return PRIORITY
